use rocket::{
    form::Form,
    get,
    http::{Cookie, CookieJar},
    post,
    request::{self, FlashMessage, FromRequest},
    response::{Flash, Redirect},
    routes,
    time::{Duration, OffsetDateTime},
    Request, Route, State,
};
use rocket_dyn_templates::{context, Template};
use sqlx::SqlitePool;
use tera::Context;

use crate::model::{
    boat::Boat,
    logbook::{
        LogToAdd, LogToFinalize, Logbook, LogbookCreateError, LogbookDeleteError,
        LogbookUpdateError,
    },
    logtype::LogType,
    user::{AdminUser, User, UserWithWaterStatus},
};

pub struct KioskCookie(String);

#[rocket::async_trait]
impl<'r> FromRequest<'r> for KioskCookie {
    type Error = std::convert::Infallible;

    async fn from_request(request: &'r Request<'_>) -> request::Outcome<KioskCookie, Self::Error> {
        match request.cookies().get_private("kiosk") {
            Some(cookie) => request::Outcome::Success(KioskCookie(cookie.value().to_string())),
            None => request::Outcome::Forward(()),
        }
    }
}

#[get("/", rank = 2)]
async fn index(
    db: &State<SqlitePool>,
    flash: Option<FlashMessage<'_>>,
    adminuser: AdminUser,
) -> Template {
    let boats = Boat::for_user(db, &adminuser.user).await;

    let coxes: Vec<UserWithWaterStatus> = futures::future::join_all(
        User::cox(db)
            .await
            .into_iter()
            .map(|user| UserWithWaterStatus::from_user(user, db)),
    )
    .await;
    let users: Vec<UserWithWaterStatus> = futures::future::join_all(
        User::all(db)
            .await
            .into_iter()
            .map(|user| UserWithWaterStatus::from_user(user, db)),
    )
    .await;
    let logtypes = LogType::all(db).await;
    let distances = Logbook::distances(db).await;

    let on_water = Logbook::on_water(db).await;

    let mut context = Context::new();
    if let Some(msg) = flash {
        context.insert("flash", &msg.into_inner());
    }

    context.insert("boats", &boats);
    context.insert("coxes", &coxes);
    context.insert("users", &users);
    context.insert("logtypes", &logtypes);
    context.insert("loggedin_user", &adminuser.user);
    context.insert("on_water", &on_water);
    context.insert("distances", &distances);

    Template::render("log", context.into_json())
}

#[get("/show", rank = 2)]
async fn show(db: &State<SqlitePool>, user: User) -> Template {
    let logs = Logbook::completed(db).await;

    Template::render("log.completed", context!(logs, loggedin_user: &user))
}

#[get("/show")]
async fn show_kiosk(db: &State<SqlitePool>, _kios: KioskCookie) -> Template {
    let logs = Logbook::completed(db).await;

    Template::render("log.completed", context!(logs))
}

#[get("/kiosk/ekrv2019/<loc>")]
fn new_kiosk(cookies: &CookieJar<'_>, loc: String) -> Redirect {
    let mut cookie = Cookie::new("kiosk", loc);
    cookie.set_expires(OffsetDateTime::now_utc() + Duration::weeks(12));
    cookies.add_private(cookie);
    Redirect::to("/log")
}

#[get("/")]
async fn kiosk(
    db: &State<SqlitePool>,
    flash: Option<FlashMessage<'_>>,
    kiosk: KioskCookie,
) -> Template {
    let boats = Boat::all_at_location(db, kiosk.0).await;
    let coxes: Vec<UserWithWaterStatus> = futures::future::join_all(
        User::cox(db)
            .await
            .into_iter()
            .map(|user| UserWithWaterStatus::from_user(user, db)),
    )
    .await;
    let users: Vec<UserWithWaterStatus> = futures::future::join_all(
        User::all(db)
            .await
            .into_iter()
            .map(|user| UserWithWaterStatus::from_user(user, db)),
    )
    .await;
    let logtypes = LogType::all(db).await;
    let distances = Logbook::distances(db).await;

    let on_water = Logbook::on_water(db).await;

    let mut context = Context::new();
    if let Some(msg) = flash {
        context.insert("flash", &msg.into_inner());
    }

    context.insert("boats", &boats);
    context.insert("coxes", &coxes);
    context.insert("users", &users);
    context.insert("logtypes", &logtypes);
    context.insert("on_water", &on_water);
    context.insert("distances", &distances);

    Template::render("kiosk", context.into_json())
}

async fn create_logbook(db: &SqlitePool, data: Form<LogToAdd>) -> Flash<Redirect> {
    match Logbook::create(
        db,
        data.into_inner()
    )
    .await
    {
        Ok(_) => Flash::success(Redirect::to("/log"), "Ausfahrt erfolgreich hinzugefügt"),
        Err(LogbookCreateError::BoatAlreadyOnWater) => Flash::error(Redirect::to("/log"), format!("Boot schon am Wasser")),
        Err(LogbookCreateError::ShipmasterAlreadyOnWater) => Flash::error(Redirect::to("/log"), format!("Schiffsführer schon am Wasser")),
        Err(LogbookCreateError::RowerAlreadyOnWater(rower)) => Flash::error(Redirect::to("/log"), format!("Ruderer {} schon am Wasser", rower.name)),
        Err(LogbookCreateError::BoatLocked) => Flash::error(Redirect::to("/log"), format!("Boot gesperrt")),
        Err(LogbookCreateError::BoatNotFound) => Flash::error(Redirect::to("/log"), format!("Boot gibt's ned")),
        Err(LogbookCreateError::TooManyRowers(expected, actual)) => Flash::error(Redirect::to("/log"), format!("Zu viele Ruderer (Boot fasst maximal {expected}, es wurden jedoch {actual} Ruderer ausgewählt)")),
        Err(LogbookCreateError::RowerCreateError(rower, e)) => Flash::error(Redirect::to("/log"), format!("Fehler bei Ruderer {rower}: {e}")),
        Err(LogbookCreateError::SamePersonShipmasterAndRower) => Flash::error(Redirect::to("/log"), format!("Selbe Person als Schiffsführer und Ruderer ausgewählt")),
        Err(LogbookCreateError::ArrivalSetButNoDistance) => Flash::error(Redirect::to("/log"), format!("Distanz notwendig, wenn Ankunftszeit angegeben wurde")),
        Err(LogbookCreateError::ArrivalSetButNoDestination) => Flash::error(Redirect::to("/log"), format!("Ziel notwendig, wenn Ankunftszeit angegeben wurde")),
        Err(LogbookCreateError::ArrivalNotAfterDeparture) => Flash::error(Redirect::to("/log"), format!("Ankunftszeit kann nicht vor der Abfahrtszeit sein")),

    }
}

#[post("/", data = "<data>", rank = 2)]
async fn create(
    db: &State<SqlitePool>,
    data: Form<LogToAdd>,
    _adminuser: AdminUser,
) -> Flash<Redirect> {
    create_logbook(db, data).await
}

#[post("/", data = "<data>")]
async fn create_kiosk(
    db: &State<SqlitePool>,
    data: Form<LogToAdd>,
    _kiosk: KioskCookie,
) -> Flash<Redirect> {
    create_logbook(db, data).await
}

async fn home_logbook(
    db: &SqlitePool,
    data: Form<LogToFinalize>,
    logbook_id: i32,
    user: &User,
) -> Flash<Redirect> {
    let logbook: Option<Logbook> = Logbook::find_by_id(db, logbook_id).await;
    let Some(logbook) = logbook else {
        return Flash::error(
            Redirect::to("/admin/log"),
            format!("Log with ID {} does not exist!", logbook_id),
        );
    };

    match logbook.home(db, user, data.into_inner()).await {
        Ok(_) => Flash::success(Redirect::to("/log"), "Successfully updated log"),
        Err(LogbookUpdateError::TooManyRowers(expected, actual)) => Flash::error(Redirect::to("/log"), format!("Zu viele Ruderer (Boot fasst maximal {expected}, es wurden jedoch {actual} Ruderer ausgewählt)")),
        Err(a) => {println!("{a:?}"); return Flash::error(
            Redirect::to("/log"),
            format!("Logbook with ID {} could not be updated!", logbook_id),
        )},
    }
}

#[post("/<logbook_id>", data = "<data>")]
async fn home_kiosk(
    db: &State<SqlitePool>,
    data: Form<LogToFinalize>,
    logbook_id: i32,
    _kiosk: KioskCookie,
) -> Flash<Redirect> {
    let logbook = Logbook::find_by_id(db, logbook_id).await.unwrap(); //TODO: fixme
    home_logbook(
        db,
        data,
        logbook_id,
        &User::find_by_id(db, logbook.shipmaster as i32)
            .await
            .unwrap(),
    )
    .await
}

#[post("/<logbook_id>", data = "<data>", rank = 2)]
async fn home(
    db: &State<SqlitePool>,
    data: Form<LogToFinalize>,
    logbook_id: i32,
    adminuser: AdminUser,
) -> Flash<Redirect> {
    home_logbook(db, data, logbook_id, &adminuser.user).await
}

#[get("/<logbook_id>/delete")]
async fn delete(db: &State<SqlitePool>, logbook_id: i32, user: User) -> Flash<Redirect> {
    let logbook = Logbook::find_by_id(db, logbook_id).await;
    if let Some(logbook) = logbook {
        match logbook.delete(db, &user).await {
            Ok(_) => Flash::success(
                Redirect::to("/log"),
                format!("Logbook with ID {} successfully deleted!", logbook_id),
            ),
            Err(LogbookDeleteError::NotYourEntry) => Flash::error(
                Redirect::to("/log"),
                "Du hast nicht die Berechtigung, den Eintrag zu löschen!",
            ),
        }
    } else {
        Flash::error(
            Redirect::to("/log"),
            format!("Logbook with ID {} could not be found!", logbook_id),
        )
    }
}

pub fn routes() -> Vec<Route> {
    routes![
        index,
        create,
        create_kiosk,
        home,
        kiosk,
        home_kiosk,
        new_kiosk,
        show,
        show_kiosk,
        delete
    ]
}

#[cfg(test)]
mod test {
    use rocket::http::ContentType;
    use rocket::{http::Status, local::asynchronous::Client};
    use sqlx::SqlitePool;

    use crate::testdb;

    #[sqlx::test]
    fn test_kiosk_cookie() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let req = client.get("/log");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::SeeOther);
        assert_eq!(response.headers().get("Location").next(), Some("/auth"));

        let req = client.get("/log/kiosk/ekrv2019/Linz");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::SeeOther);
        assert_eq!(response.headers().get("Location").next(), Some("/log"));

        let req = client.get("/log");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::Ok);
        let text = response.into_string().await.unwrap();
        assert!(text.contains("Logbuch"));
        assert!(text.contains("Neue Ausfahrt"));

        assert!(!text.contains("Ottensheim Boot"));
    }

    #[sqlx::test]
    fn test_kiosk_cookie_boat() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let req = client.get("/log/kiosk/ekrv2019/Ottensheim");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::SeeOther);
        assert_eq!(response.headers().get("Location").next(), Some("/log"));

        let req = client.get("/log");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::Ok);
        let text = response.into_string().await.unwrap();
        assert!(text.contains("Logbuch"));
        assert!(text.contains("Neue Ausfahrt"));

        assert!(text.contains("Ottensheim Boot"));
    }

    #[sqlx::test]
    fn test_index() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let login = client
            .post("/auth")
            .header(ContentType::Form) // Set the content type to form
            .body("name=admin&password=admin"); // Add the form data to the request body;
        login.dispatch().await;

        let req = client.get("/log");
        let response = req.dispatch().await;

        let text = response.into_string().await.unwrap();
        assert!(text.contains("Logbuch"));
        assert!(text.contains("Neue Ausfahrt"));
    }

    #[sqlx::test]
    fn test_show() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let login = client
            .post("/auth")
            .header(ContentType::Form) // Set the content type to form
            .body("name=admin&password=admin"); // Add the form data to the request body;
        login.dispatch().await;

        let req = client.get("/log/show");
        let response = req.dispatch().await;

        let text = response.into_string().await.unwrap();
        println!("{text:?}");
        assert!(text.contains("Logbuch"));
        assert!(text.contains("Joe"));
    }

    #[sqlx::test]
    fn test_show_kiosk() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let req = client.get("/log/kiosk/ekrv2019/Linz");
        let _ = req.dispatch().await;

        let req = client.get("/log/show");
        let response = req.dispatch().await;

        let text = response.into_string().await.unwrap();
        assert!(text.contains("Logbuch"));
        assert!(text.contains("Joe"));
    }

    #[sqlx::test]
    fn test_create() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let login = client
            .post("/auth")
            .header(ContentType::Form) // Set the content type to form
            .body("name=admin&password=admin"); // Add the form data to the request body;
        login.dispatch().await;

        let req = client
            .post("/log")
            .header(ContentType::Form)
            .body("boat_id=1&shipmaster=4&departure=2199-12-31T10:00");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::SeeOther);
        assert_eq!(response.headers().get("Location").next(), Some("/log"));

        let flash_cookie = response
            .cookies()
            .get("_flash")
            .expect("Expected flash cookie");

        assert_eq!(
            flash_cookie.value(),
            "7:successAusfahrt erfolgreich hinzugefügt"
        );
    }

    #[sqlx::test]
    fn test_home_kiosk() {
        let db = testdb!();

        let rocket = rocket::build().manage(db.clone());
        let rocket = crate::tera::config(rocket);

        let client = Client::tracked(rocket).await.unwrap();
        let req = client.get("/log/kiosk/ekrv2019/Linz");
        let _ = req.dispatch().await;

        let req = client
            .post("/log/1")
            .header(ContentType::Form)
            .body("destination=Ottensheim&distance_in_km=25");
        let response = req.dispatch().await;

        assert_eq!(response.status(), Status::SeeOther);
        assert_eq!(response.headers().get("Location").next(), Some("/log"));

        let flash_cookie = response
            .cookies()
            .get("_flash")
            .expect("Expected flash cookie");

        assert_eq!(flash_cookie.value(), "7:successSuccessfully updated log");
    }
}
